@@ -1,33 +1,38 @@
 # Copyright: 2011, Grigoriy Petukhov
 # Author: Grigoriy Petukhov (http://lorien.name)
 # License: BSD
-from __future__ import absolute_import
 import email
 import logging
-import urllib
-from StringIO import StringIO
+#import urllib
+#try:
+#    from StringIO import StringIO
+#except ImportError:
+#    from io import StringIO
 import threading
 import random
+import requests 
 
-from ..base import GrabError, GrabMisuseError, UploadContent, UploadFile
+from grab.error import GrabError, GrabMisuseError
+from grab.base import UploadContent, UploadFile
+from grab.response import Response
+from grab.tools.http import urlencode, normalize_http_values, normalize_unicode
+from grab.util.py3k_support import *
 
-logger = logging.getLogger('grab')
-
-class RequestsTransportExtension(object):
-    def extra_init(self):
-        import requests 
+logger = logging.getLogger('grab.transport.requests')
 
+class RequestsTransport(object):
+    def __init__(self):
         self.session = requests.session()
 
-    def extra_reset(self):
-        #self.response_head_chunks = []
-        #self.response_body_chunks = []
-        #self.response_body_bytes_read = 0
+    def reset(self):
+        # TODO: WTF???
+        # Maybe move to super-class???
         self.request_headers = ''
         self.request_head = ''
-        self.request_log = ''
         self.request_body = ''
-        self.request_method = None
+        self.request_log = ''
+
+        #self.request_method = None
         self.requests_config = None
 
     #def head_processor(self, chunk):
@@ -35,7 +40,7 @@ class RequestsTransportExtension(object):
         #Process head of response.
         #"""
 
-        #if self.config['nohead']:
+        #if grab.config['nohead']:
             #return 0
         #self.response_head_chunks.append(chunk)
         ## Returning None implies that all bytes were written
@@ -46,7 +51,7 @@ class RequestsTransportExtension(object):
         #Process body of response.
         #"""
 
-        #if self.config['nobody']:
+        #if grab.config['nobody']:
             #return 0
         #self.response_body_chunks.append(chunk)
         ## Returning None implies that all bytes were written
@@ -78,46 +83,50 @@ class RequestsTransportExtension(object):
                 #self.request_log = ''
             #self.request_log += text
 
-    def process_config(self):
+    def process_config(self, grab):
         """
-        Setup curl instance with values from ``self.config``.
+        Setup curl instance with values from ``grab.config``.
         """
         
         # Accumulate all request options into `self.requests_config`
-        self.requests_config = {'headers': {}, 'payload': None,
-                                'cookies': None, 'proxy': None}
+        self.requests_config = {
+            'headers': {},
+            'payload': None,
+            'cookies': None,
+            'proxy': None,
+        }
 
-        if isinstance(self.config['url'], unicode):
-            self.config['url'] = self.config['url'].encode('utf-8')
+        if isinstance(grab.config['url'], unicode):
+            grab.config['url'] = grab.config['url'].encode('utf-8')
 
-        self.requests_config['url'] = self.config['url']
+        self.requests_config['url'] = grab.config['url']
 
         #self.curl.setopt(pycurl.URL, url)
         #self.curl.setopt(pycurl.FOLLOWLOCATION, 1)
         #self.curl.setopt(pycurl.MAXREDIRS, 5)
-        #self.curl.setopt(pycurl.CONNECTTIMEOUT, self.config['connect_timeout'])
-        #self.curl.setopt(pycurl.TIMEOUT, self.config['timeout'])
+        #self.curl.setopt(pycurl.CONNECTTIMEOUT, grab.config['connect_timeout'])
+        #self.curl.setopt(pycurl.TIMEOUT, grab.config['timeout'])
         #self.curl.setopt(pycurl.NOSIGNAL, 1)
         #self.curl.setopt(pycurl.WRITEFUNCTION, self.body_processor)
         #self.curl.setopt(pycurl.HEADERFUNCTION, self.head_processor)
 
         # User-Agent
         # TODO: move to base class
-        if self.config['user_agent'] is None:
-            if self.config['user_agent_file'] is not None:
-                lines = open(self.config['user_agent_file']).read().splitlines()
-                self.config['user_agent'] = random.choice(lines)
+        if grab.config['user_agent'] is None:
+            if grab.config['user_agent_file'] is not None:
+                lines = open(grab.config['user_agent_file']).read().splitlines()
+                grab.config['user_agent'] = random.choice(lines)
 
         # If value is None then set empty string
         # None is not acceptable because in such case
         # pycurl will set its default user agent "PycURL/x.xx.x"
         # For consistency we send empty User-Agent in case of None value
         # in all other transports too
-        if not self.config['user_agent']:
-            self.config['user_agent'] = ''
-        self.requests_config['headers']['User-Agent'] = self.config['user_agent']
+        if not grab.config['user_agent']:
+            grab.config['user_agent'] = ''
+        self.requests_config['headers']['User-Agent'] = grab.config['user_agent']
 
-        #if self.config['debug']:
+        #if grab.config['debug']:
             #self.curl.setopt(pycurl.VERBOSE, 1)
             #self.curl.setopt(pycurl.DEBUGFUNCTION, self.debug_processor)
 
@@ -125,35 +134,36 @@ class RequestsTransportExtension(object):
         #self.curl.setopt(pycurl.SSL_VERIFYPEER, 0)
         #self.curl.setopt(pycurl.SSL_VERIFYHOST, 0)
 
-        self.requests_config['method'] = self.request_method.lower()
+        self.requests_config['method'] = grab.request_method.lower()
 
-        if self.request_method == 'POST' or self.request_method == 'PUT':
-            if self.config['multipart_post']:
+        if grab.request_method == 'POST' or grab.request_method == 'PUT':
+            if grab.config['multipart_post']:
                 raise NotImplementedError
-                #if isinstance(self.config['multipart_post'], basestring):
+                #if isinstance(grab.config['multipart_post'], basestring):
                     #raise GrabMisuseError('multipart_post option could not be a string')
-                #post_items = self.normalize_http_values(self.config['multipart_post'])
+                #post_items = normalize_http_values(grab.config['multipart_post'],
+                                                    #charset=grab.config['charset'])
                 #self.curl.setopt(pycurl.HTTPPOST, post_items) 
-            elif self.config['post']:
-                if isinstance(self.config['post'], basestring):
+            elif grab.config['post']:
+                if isinstance(grab.config['post'], basestring):
                     # bytes-string should be posted as-is
                     # unicode should be converted into byte-string
-                    if isinstance(self.config['post'], unicode):
-                        post_data = self.normalize_unicode(self.config['post'])
+                    if isinstance(grab.config['post'], unicode):
+                        post_data = normalize_unicode(grab.config['post'])
                     else:
-                        post_data = self.config['post']
+                        post_data = grab.config['post']
                 else:
                     # dict, tuple, list should be serialized into byte-string
-                    post_data = self.urlencode(self.config['post'])
+                    post_data = urlencode(grab.config['post'])
                 self.requests_config['payload'] = post_data
                 #self.curl.setopt(pycurl.POSTFIELDS, post_data)
-        #elif self.request_method == 'PUT':
+        #elif grab.request_method == 'PUT':
             #self.curl.setopt(pycurl.PUT, 1)
-            #self.curl.setopt(pycurl.READFUNCTION, StringIO(self.config['post']).read) 
-        elif self.request_method == 'DELETE':
+            #self.curl.setopt(pycurl.READFUNCTION, StringIO(grab.config['post']).read) 
+        elif grab.request_method == 'DELETE':
             pass
             #self.curl.setopt(pycurl.CUSTOMREQUEST, 'delete')
-        elif self.request_method == 'HEAD':
+        elif grab.request_method == 'HEAD':
             pass
             #self.curl.setopt(pycurl.NOBODY, 1)
         else:
@@ -161,56 +171,54 @@ class RequestsTransportExtension(object):
             #self.curl.setopt(pycurl.HTTPGET, 1)
 
         
-        headers = self.default_headers
-        if self.config['headers']:
-            headers.update(self.config['headers'])
+        headers = grab.config['common_headers']
+        if grab.config['headers']:
+            headers.update(grab.config['headers'])
         #header_tuples = [str('%s: %s' % x) for x\
                          #in headers.iteritems()]
         #self.curl.setopt(pycurl.HTTPHEADER, header_tuples)
         self.requests_config['headers'].update(headers)
 
-        if self.config['cookies']:
-            items = self.normalize_http_values(self.config['cookies'])
+        # `cookiefile` option shoul be processed before `cookies` option
+        # because `load_cookies` updates `cookies` option
+        if grab.config['cookiefile']:
+            grab.load_cookies(grab.config['cookiefile'])
+
+        if grab.config['cookies']:
+            items = normalize_http_values(grab.config['cookies'])
             self.requests_config['cookies'] = dict(items)
 
-        #if not self.config['reuse_cookies'] and not self.config['cookies']:
+        #if not grab.config['reuse_cookies'] and not grab.config['cookies']:
             #self.curl.setopt(pycurl.COOKIELIST, 'ALL')
 
-        if self.config['cookiefile']:
-            self.load_cookies(self.config['cookiefile'])
+        #if grab.config['referer']:
+            #self.curl.setopt(pycurl.REFERER, str(grab.config['referer']))
 
-
-        #if self.config['referer']:
-            #self.curl.setopt(pycurl.REFERER, str(self.config['referer']))
-
-        #if self.config['proxy']:
-            #self.curl.setopt(pycurl.PROXY, str(self.config['proxy'])) 
+        #if grab.config['proxy']:
+            #self.curl.setopt(pycurl.PROXY, str(grab.config['proxy'])) 
         #else:
             #self.curl.setopt(pycurl.PROXY, '')
 
-        #if self.config['proxy_userpwd']:
-            #self.curl.setopt(pycurl.PROXYUSERPWD, self.config['proxy_userpwd'])
+        #if grab.config['proxy_userpwd']:
+            #self.curl.setopt(pycurl.PROXYUSERPWD, grab.config['proxy_userpwd'])
 
-        if self.config['proxy']:
-            self.requests_config['proxy'] = self.config['proxy']
+        if grab.config['proxy']:
+            self.requests_config['proxy'] = grab.config['proxy']
 
-        if self.config['proxy_userpwd']:
+        if grab.config['proxy_userpwd']:
             raise GrabMisuseError('requests transport does not support proxy authentication')
 
-        if self.config['proxy_type']:
-            if self.config['proxy_type'] != 'http':
+        if grab.config['proxy_type']:
+            if grab.config['proxy_type'] != 'http':
                 raise GrabMisuseError('requests transport supports only proxies of http type')
 
-        #if self.config['encoding']:
-            #self.curl.setopt(pycurl.ENCODING, self.config['encoding'])
+        #if grab.config['encoding']:
+            #self.curl.setopt(pycurl.ENCODING, grab.config['encoding'])
 
-        #if self.config['userpwd']:
-            #self.curl.setopt(pycurl.USERPWD, self.config['userpwd'])
+        #if grab.config['userpwd']:
+            #self.curl.setopt(pycurl.USERPWD, grab.config['userpwd'])
 
-        #if self.config['charset']:
-            #self.charset = self.config['charset']
-
-    #def extract_cookies(self):
+    #def _extract_cookies(self):
         #"""
         #Extract cookies.
         #"""
@@ -224,8 +232,7 @@ class RequestsTransportExtension(object):
         #return cookies
 
 
-    def transport_request(self):
-        import requests
+    def request(self):
         try:
             cfg = self.requests_config
             func = getattr(requests, cfg['method'])
@@ -239,23 +246,37 @@ class RequestsTransportExtension(object):
                                      'https': cfg['proxy']}
             self._requests_response = func(
                 cfg['url'], headers=cfg['headers'], **kwargs)
-        except Exception, ex:
+        except Exception as ex:
             raise GrabError(0, unicode(ex))
 
-    def prepare_response(self):
+    def prepare_response(self, grab):
         #self.response.head = ''.join(self.response_head_chunks)
         #self.response.body = ''.join(self.response_body_chunks)
         #self.response.parse()
-        #self.response.cookies = self.extract_cookies()
+        #self.response.cookies = self._extract_cookies()
         #self.response.code = self.curl.getinfo(pycurl.HTTP_CODE)
         #self.response.time = self.curl.getinfo(pycurl.TOTAL_TIME)
         #self.response.url = self.curl.getinfo(pycurl.EFFECTIVE_URL)
-        self.response.head = ''
-        self.response.body = self._requests_response.content
-        self.response.code = self._requests_response.status_code
-        self.response.headers = self._requests_response.headers
-        self.response.cookies = self._requests_response.cookies or {}
-        self.response.url = self.config['url']
+        response = Response()
+        response.head = ''
+        #if grab.config['body_max_size'] is not None:
+            #chunks = []
+            #read_size = 0
+            #for chunk in self._requests_responsek
+        #else:
+            #response.body = self._requests_response.content
+
+        response.body = self._requests_response.content
+        response.code = self._requests_response.status_code
+        response.headers = self._requests_response.headers
+        response.cookies = self._requests_response.cookies or {}
+        response.url = grab.config['url']
+
+        if grab.config['charset'] is not None:
+            response.parse(charset=grab.config['charset'])
+        else:
+            response.parse()
+        return response
 
     #def load_cookies(self, path):
         #"""
@@ -284,7 +305,7 @@ class RequestsTransportExtension(object):
 
         ## Write tests
         #self.curl.setopt(pycurl.COOKIELIST, 'ALL')
-        #self.config['cookies'] = {}
+        #grab.config['cookies'] = {}
         #self.response.cookies = None
 
     #def reset_curl_instance(self):
@@ -299,6 +320,6 @@ class RequestsTransportExtension(object):
         #self.curl = pycurl.Curl()
 
 
-from ..base import BaseGrab
-class GrabRequests(RequestsTransportExtension, BaseGrab):
-    pass
+#from grab.base import BaseGrab
+#class GrabRequests(RequestsTransportExtension, BaseGrab):
+    #pass
